﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Migrations;
using ClientServer.Models.Users;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Controllers.Core.Users
{
    [Route(Constants.ApiPrefix + "groups")]
    public class GroupController : ControllerWithDb
    {
        public GroupController(YapexDbContext context) : base(context)
        {
        }

        /// <summary>
        /// creates a new group and adds the user as creator
        /// </summary>
        /// <param name="groupFromFrontend">the group to create</param>
        [HttpPost("create")]
        public async Task AddGroup([FromBody] UserGroupForBackend groupFromFrontend)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            if (
                !await HasSystemPermission(
                    permission => permission != null && permission.SystemRolePermission.CanCreateGroups))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            //check if user really exists

            User user = await _context.Users.FirstOrDefaultAsync(p => p.Id == userId);

            if (user == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user not found")));
                return;
            }

            SystemSetting systemSetting = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSetting == null)
            {
                Console.WriteLine("ERROR: system settings not found");
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "system settings not found")));
                return;
            }

            var group = new UserGroup()
            {
                DisplayName = groupFromFrontend.DisplayName.Trim(),
                Email = groupFromFrontend.Email.Trim()
            };

            var connection = new UserWithUserGroup()
            {
                User = user,
                GroupRoleId = systemSetting.DefaultGroupCreatorGroupRoleId,
                UserGroup = group
            };

            try
            {
                _context.UserWithUserGroups.Add(connection);
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            var userGroupForFrontend = new UserGroupBase()
            {
                Id = group.Id,
                DisplayName = group.DisplayName,
                Email = group.Email
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<UserGroupBase>(ResponseCode.Ok, "user group created",
                        userGroupForFrontend)));
        }

        /// <summary>
        /// changes the group data
        /// </summary>
        /// <param name="group">the group</param>
        [HttpPost("update")]
        public async Task Change([FromBody] UserGroupForBackend group)
        {
            if (!await base.IsLoggedIn()) return;

            if (!await HasGroupPermission(
                group.Id,
                permission => permission != null && permission.CanChangeGroupData)
            )
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            UserGroup oldGroup = await _context.UserGroups
                .FirstOrDefaultAsync(p => p.Id == group.Id);

            if (oldGroup == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user group not found")));
                return;
            }

            oldGroup.DisplayName = group.DisplayName.Trim();
            oldGroup.Email = group.Email.Trim();


            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            var groupForFrontend = new UserGroupBase()
            {
                Id = oldGroup.Id,
                DisplayName = oldGroup.DisplayName,
                Email = oldGroup.Email
            };


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<UserGroupBase>(ResponseCode.Ok, "user group changed",
                        groupForFrontend)));
        }


        /// <summary>
        /// deletes the given user group
        /// </summary>
        /// <param name="groupId">the user group id</param>
        [HttpDelete("delete/{groupId}")]
        public async Task DeleteGroup(int groupId)
        {
            if (!await base.IsLoggedIn()) return;

            if (
                !await HasSystemPermission(
                    permission => permission != null && permission.SystemRolePermission.CanDeleteGroups))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            SystemSetting systemSetting = await _context.SystemSettings.FirstOrDefaultAsync();

            if (systemSetting == null)
            {
                Console.WriteLine("ERROR: system settings not found");
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "system settings not found")));
                return;
            }

            UserGroup group = await _context.UserGroups.FirstOrDefaultAsync(p => p.Id == groupId);

            if (group == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "group not found")));
                return;
            }

            //IsOnlyVisibleToOwner == true --> there are only releases for the owner that can be deleted
            var exercisesInGroup = await _context.Exercises
                    .CountAsync(p => p.UserGroupId == group.Id && p.IsOnlyVisibleToOwner == false)
                ;

            if (exercisesInGroup > 0)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "group has still exercises")));
                return;
            }

            //cannot delete default group
            if (group.Id == systemSetting.DefaultUserGroupId)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "cannot delete the default user group (system settings)")));
                return;
            }


            try
            {
                _context.UserGroups.Remove(group);
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "group deleted")));
        }


        /// <summary>
        /// returns all groups where the current user is a member
        /// </summary>
        [HttpPost("fromuser")]
        public async Task GetGroupsFromUser([FromBody] PaginationPostData pagination)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            var connections = _context.UserWithUserGroups
                .Include(p => p.UserGroup)
                .Include(p => p.GroupRole)
                .ThenInclude(p => p.GroupRolePermission)
                .Where(p => p.UserId == userId);


            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                connections = connections
                        .Where(p => p.UserGroup != null &&
                                    (
                                        (p.UserGroup.DisplayName != null &&
                                         p.UserGroup.DisplayName.ToLower().Contains(text))
                                        || (p.UserGroup.Email != null && p.UserGroup.Email.ToLower().Contains(text))
                                    )
                        )
                    ;
            }

            string orderByProp = "";
            Expression<Func<UserWithUserGroup, object>> orderByFunc = null;

            //not all props are sortable via the frontend
            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "id".ToLower())
                {
                    orderByFunc = p => p.UserGroup.Id;
                    orderByProp = "id".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "displayName".ToLower())
                {
                    orderByFunc = p => p.UserGroup.DisplayName;
                    orderByProp = "displayName";
                }
                else if (pagination.OrderByProp.ToLower() == "email".ToLower())
                {
                    orderByFunc = p => p.UserGroup.Email;
                    orderByProp = "email";
                }
            }

            var paginatedData = await PaginationHelper.GetAsPaginationAsync(connections,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );

            var groups = new List<UserGroupWithExerciseCountFromBackend>();

            foreach (var userWithUserGroup in paginatedData.Items)
            {
                //IsOnlyVisibleToOwner == true --> there are only releases for the owner that can be deleted
                var exercisesInGroup = await _context.Exercises
                        .CountAsync(p => p.UserGroupId == userWithUserGroup.UserGroupId && p.IsOnlyVisibleToOwner == false)
                    ;
                
                groups.Add(new UserGroupWithExerciseCountFromBackend()
                {
                    Id = userWithUserGroup.UserGroup.Id,
                    DisplayName = userWithUserGroup.UserGroup.DisplayName,
                    Email = userWithUserGroup.UserGroup.Email,
                    GroupRole = new GroupRoleWithPermissionFullBase
                    (
                        userWithUserGroup.GroupRole.Id,
                        userWithUserGroup.GroupRole.DisplayName,
                        canChangeExercises: userWithUserGroup.GroupRole.GroupRolePermission.CanChangeExercises,
                        canCreateExercises: userWithUserGroup.GroupRole.GroupRolePermission.CanCreateExercises,
                        canAssessExercises: userWithUserGroup.GroupRole.GroupRolePermission.CanAssessExercises,
                        canManageExerciseReleases: userWithUserGroup.GroupRole.GroupRolePermission.CanManageExerciseReleases,
                        canDeleteExercises: userWithUserGroup.GroupRole.GroupRolePermission.CanDeleteExercises,
                        canAddUserToGroup: userWithUserGroup.GroupRole.GroupRolePermission.CanAddUserToGroup,
                        canLockExercisesPermanently: userWithUserGroup.GroupRole.GroupRolePermission.CanLockExercisesPermanently,
                        canSeeExercisesFromOthersInGroup: userWithUserGroup.GroupRole.GroupRolePermission
                            .CanSeeExercisesFromOthersInGroup,
                        canChangeOtherMembersRole:
                        userWithUserGroup.GroupRole.GroupRolePermission.CanChangeOtherMembersRole,
                        canRemoveMemberFromGroup:
                        userWithUserGroup.GroupRole.GroupRolePermission.CanRemoveMemberFromGroup,
                        canSeeOtherMembers: userWithUserGroup.GroupRole.GroupRolePermission.CanSeeOtherMembers,
                        canChangeGroupData: userWithUserGroup.GroupRole.GroupRolePermission.CanChangeGroupData
                    ),
                    NumExercisesInGroup = exercisesInGroup
                });
            }

            var frontendData = PaginationHelper.TransformForFrontend(paginatedData, groups);

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<UserGroupWithExerciseCountFromBackend>>(ResponseCode.Ok,
                        "groups queried",
                        frontendData)));
        }


        /// <summary>
        /// returns a single group where the user is a member
        /// </summary>
        /// <param name="groupId"></param>
        /// <returns></returns>
        [HttpGet("fromuser/{groupId}")]
        public async Task GetSingleGroup(int groupId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            var connection = await _context.UserWithUserGroups
                    .Include(p => p.UserGroup)
                    .Include(p => p.GroupRole)
                    .ThenInclude(p => p.GroupRolePermission)
                    .FirstOrDefaultAsync(p => p.UserId == userId && p.UserGroupId == groupId)
                ;

            if (connection == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "user group not found")));
                return;
            }
            
            //IsOnlyVisibleToOwner == true --> there are only releases for the owner that can be deleted
            var exercisesInGroup = await _context.Exercises
                    .CountAsync(p => p.UserGroupId == connection.UserGroupId && p.IsOnlyVisibleToOwner == false)
                ;

            var frontendData = new UserGroupWithExerciseCountFromBackend()
            {
                Id = connection.UserGroup.Id,
                DisplayName = connection.UserGroup.DisplayName,
                Email = connection.UserGroup.Email,
                GroupRole =
                    new
                        GroupRoleWithPermissionFullBase //get the group role to know e.g. if we can delete users or not
                        (
                            id: connection.GroupRole.Id,
                            displayName: connection.GroupRole.DisplayName,
                            canChangeExercises: connection.GroupRole.GroupRolePermission.CanChangeExercises,
                            canCreateExercises: connection.GroupRole.GroupRolePermission.CanCreateExercises,
                            canAssessExercises: connection.GroupRole.GroupRolePermission.CanAssessExercises,
                            canManageExerciseReleases: connection.GroupRole.GroupRolePermission.CanManageExerciseReleases,
                            canDeleteExercises: connection.GroupRole.GroupRolePermission.CanDeleteExercises,
                            canAddUserToGroup: connection.GroupRole.GroupRolePermission.CanAddUserToGroup,
                            canLockExercisesPermanently: connection.GroupRole.GroupRolePermission.CanLockExercisesPermanently,
                            canSeeExercisesFromOthersInGroup: connection.GroupRole.GroupRolePermission
                                .CanSeeExercisesFromOthersInGroup,
                            canChangeOtherMembersRole:
                            connection.GroupRole.GroupRolePermission.CanChangeOtherMembersRole,
                            canRemoveMemberFromGroup:
                            connection.GroupRole.GroupRolePermission.CanRemoveMemberFromGroup,
                            canSeeOtherMembers: connection.GroupRole.GroupRolePermission.CanSeeOtherMembers,
                            canChangeGroupData: connection.GroupRole.GroupRolePermission.CanChangeGroupData
                        ),
                NumExercisesInGroup = exercisesInGroup
            };

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<UserGroupWithExerciseCountFromBackend>(ResponseCode.Ok, "groups queried",
                        frontendData)));
        }

        /// <summary>
        /// returns the users for the given group and the group
        /// </summary>
        /// <param name="groupId">the group</param>
        [HttpPost("{groupId}/members")]
        public async Task GetMembers(int groupId, [FromBody] PaginationPostData pagination)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            

            //user needs to be member of the group to see the members...
            UserWithUserGroup connection = await _context.UserWithUserGroups.FirstOrDefaultAsync(p =>
                p.UserId == userId && p.UserGroupId == groupId);

            if (connection == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "user group not found")));
                return;
            }
            
            if (
                !await HasGroupPermission(groupId,
                    permission => permission != null && permission.CanSeeOtherMembers))
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            var userWithGroupConenction = _context.UserWithUserGroups
                //group role id
                .Include(p => p.User)
                .Include(p => p.GroupRole)
                .Where(p => p.UserGroup.Id == groupId);


            if (String.IsNullOrWhiteSpace(pagination.SearchText) == false)
            {
                var text = pagination.SearchText.ToLower();
                userWithGroupConenction = userWithGroupConenction
                        .Where(p => p.User.FirstName.ToLower().Contains(text)
                                    || p.User.LastName.ToLower().Contains(text)
                                    || p.User.Email.ToLower().Contains(text)
                                    || p.User.Token.ToLower().Contains(text)
                                    || (p.GroupRole != null && p.GroupRole.DisplayName.ToLower().Contains(text))
                        )
                    ;
            }

            string orderByProp = "";
            Expression<Func<UserWithUserGroup, object>> orderByFunc = null;

            //not all props are sortable via the frontend
            if (String.IsNullOrWhiteSpace(pagination.OrderByProp) == false)
            {
                if (pagination.OrderByProp.ToLower() == "id".ToLower())
                {
                    orderByFunc = p => p.User.Id;
                    orderByProp = "id".ToLower();
                }
                else if (pagination.OrderByProp.ToLower() == "LastName".ToLower())
                {
                    orderByFunc = p => p.User.LastName;
                    orderByProp = "lastName";
                }
                else if (pagination.OrderByProp.ToLower() == "FirstName".ToLower())
                {
                    orderByFunc = p => p.User.FirstName;
                    orderByProp = "firstName";
                }
                else if (pagination.OrderByProp.ToLower() == "Token".ToLower())
                {
                    orderByFunc = p => p.User.Token;
                    orderByProp = "token";
                }
                else if (pagination.OrderByProp.ToLower() == "Email".ToLower())
                {
                    orderByFunc = p => p.User.Email;
                    orderByProp = "email";
                }
                else if (pagination.OrderByProp.ToLower() == "groupRole".ToLower())
                {
                    orderByFunc = p => p.GroupRole.DisplayName;
                    orderByProp = "groupRole";
                }
                else if (pagination.OrderByProp.ToLower() == "isActivated".ToLower())
                {
                    orderByFunc = p => p.User.IsActivated;
                    orderByProp = "isActivated";
                }
            }

            var paginatedData = await PaginationHelper.GetAsPaginationAsync(userWithGroupConenction,
                pagination.Page, pagination.PageSize, pagination.IsOrderByDesc,
                orderByFunc, orderByProp, pagination.SearchText
            );

            var responseUsers = new List<GroupUserFromBackend>();

            foreach (UserWithUserGroup userWithUserGroup in paginatedData.Items)
            {
                responseUsers.Add(new GroupUserFromBackend()
                {
                    Id = userWithUserGroup.User.Id,
                    Token = userWithUserGroup.User.Token,
                    Email = userWithUserGroup.User.Email,
                    FirstName = userWithUserGroup.User.FirstName,
                    LastName = userWithUserGroup.User.LastName,
                    IsActivated = userWithUserGroup.User.IsActivated,
                    GroupRoleId = userWithUserGroup.GroupRoleId,
                    IsCurrentUser = userWithUserGroup.UserId == userId
                });
            }

            var frontendData = PaginationHelper.TransformForFrontend(paginatedData, responseUsers);


            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<PaginatedData<GroupUserFromBackend>>(ResponseCode.Ok, "",
                        frontendData)));
        }


        /// <summary>
        /// adds multiple users to the given group
        /// </summary>
        /// <param name="tuple">the group and users</param>
        [HttpPost("addmembers")]
        public async Task AddMembers([FromBody] AddUsersToGroupTuplesForBackend tuple)
        {
            if (!await base.IsLoggedIn()) return;

            if (
                !await HasGroupPermission(tuple.GroupId,
                    permission => permission != null && permission.CanAddUserToGroup))
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }


            UserGroup group = await _context.UserGroups
                .FirstOrDefaultAsync(p => p.Id == tuple.GroupId);


            //get default role

            SystemSetting systemSetting =
                await _context.SystemSettings.Include(p => p.DefaultGroupRole).FirstOrDefaultAsync();

            if (systemSetting == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "system settings were null")));
                return;
            }

            if (group == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "group not found")));
                return;
            }

            foreach (var userId in tuple.UserIds)
            {
                User user = await _context.Users.FirstOrDefaultAsync(p => p.Id == userId && p.IsActivated);

                if (user == null)
                {
                    //do not try to add non existing user
                }
                else
                {
                    group.UserWithUserGroups.Add(new UserWithUserGroup()
                    {
                        UserId = userId,
                        GroupRole = systemSetting.DefaultGroupRole
                    });
                }
            }

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "users added")));
        }


        /// <summary>
        /// changes the role of a group member 
        /// </summary>
        /// <param name="tuple">the tuple</param>
        [HttpPost("changememberrole")]
        public async Task ChangeMemberRole([FromBody] UserWithGroupRoleForBackend tuple)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            //a user cannot degrade his/herself

            if (userId == tuple.UserId)
            {
                //avoids some mistakes...
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                        "cannot set own group role")));
                return;
            }

            if (!await HasGroupPermission(tuple.GroupId,
                p => p != null && p.CanChangeOtherMembersRole))
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return;
            }

            //check if role exists...

            GroupRole role = await _context.GroupRoles.FirstOrDefaultAsync(p => p.Id == tuple.GroupRoleId);

            if (role == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "cannot find group role")));
                return;
            }

            //this implicitely checks if the user exists...
            UserWithUserGroup connection =
                await _context.UserWithUserGroups.FirstOrDefaultAsync(
                    p => p.UserId == tuple.UserId && p.UserGroupId == tuple.GroupId);


            if (connection == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "cannot find group connection")));
                return;
            }


            //finally check íf the group have at least on creator that could handle the group if the current user dops privileges

            bool? check = await HasGroupAtLeastOnCreator(tuple.GroupId, connection.UserId, connection.GroupRoleId);

            if (tuple.UserId == userId)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, "cannot change own role")));
                return;
            }

            if (check == null)
            {
                Console.WriteLine("ERROR: no default config --> cannot add users");
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError,
                            "internal server error (no default config)")));
            }
            else if (check.Value == false)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "group needs at least on member with the creator role")));
                return;
            }


            try
            {
                connection.GroupRoleId = role.Id;
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }


            await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "changed group role")));
        }


        /// <summary>
        /// checks if there is at least one creator left in the group when the user would leave the group
        /// <para></para>
        /// (user can always leave if he/she is not in a group creator role)
        /// </summary>
        /// <param name="groupId">the group to leave</param>
        /// <param name="userId">the user who wants to leave</param>
        /// <param name="userGroupRoleId">the current role of the user in the group</param>
        /// <returns>null: system settings not found (no creator role), true: user can leave, false: cannot leave</returns>
        private async Task<bool?> HasGroupAtLeastOnCreator(int groupId, int userId, int userGroupRoleId)
        {
            SystemSetting systemSetting = await _context.SystemSettings
                .Include(p => p.DefaultGroupRole)
                .FirstOrDefaultAsync();

            if (systemSetting == null)
            {
                return null;
            }


            //we need at least one user with the default creator role...
            if (userGroupRoleId == systemSetting.DefaultGroupCreatorGroupRoleId)
            {
                //in group and also a creator but not the current user -> we have at least one other creator...
                var count = await _context.UserWithUserGroups.CountAsync(
                        p =>
                            p.UserGroupId == groupId &&
                            p.GroupRoleId == systemSetting.DefaultGroupCreatorGroupRoleId &&
                            p.UserId != userId)
                    ;


                return count > 0;
            }

            //not a creator -> can always leave
            return true;
        }

        /// <summary>
        /// determines if the given user can leave the group or not
        /// </summary>
        /// <param name="tuple">the user with the group to leave</param>
        [HttpPost("canleave")]
        public async Task CanLeaveGroup([FromBody] RemoveUsersFromGroupTuplesForBackend tuple)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            if (tuple.UseCurrentUser)
            {
                //in this case the frontend has not set the user id
                tuple.UserId = userId;
            }

            UserWithUserGroup connection = await _context.UserWithUserGroups
                .Include(p => p.GroupRole)
                .FirstOrDefaultAsync(
                    p => p.UserId == tuple.UserId && p.UserGroupId == tuple.GroupId);

            if (connection == null)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "connection not found")));
            }
            else
            {
                bool? canLeave = await HasGroupAtLeastOnCreator(tuple.GroupId, tuple.UserId, connection.GroupRoleId);

                if (canLeave.HasValue == false)
                {
                    Console.WriteLine("ERROR: system settings not found");
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "system setting not found")));
                    return;
                }

                SystemSetting systemSetting = await _context.SystemSettings
                    .FirstOrDefaultAsync();

                //HasGroupAtLeastOnCreator checks the system settings...
                if (tuple.GroupId == systemSetting.DefaultUserGroupId)
                {
                    //user cannot leave the default group...
                    await
                        Response.WriteAsync(
                            Jc.Serialize(new BasicResponseWithData<bool>(ResponseCode.Ok,
                                "user cannot leave the default group", false)));
                    return;
                }


                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<bool>(ResponseCode.Ok, "", canLeave.Value)));
            }
        }


        //TODO creator cannot be removed??...
        // -> we will always have at least one member with creator in the group (remove group)
        [HttpPost("removemember")]
        public async Task RemoveMember([FromBody] RemoveUsersFromGroupTuplesForBackend tuple)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();


            if (tuple.UseCurrentUser)
            {
                //in this case the frontend has not set the user id
                tuple.UserId = userId;
            }

            //user has always the permission to remove him/herself from a group...
            if (userId != tuple.UserId
                && !await HasGroupPermission(tuple.GroupId,
                    permission => permission != null && permission.CanRemoveMemberFromGroup))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NoPermission,
                            "no permission to remove other user")));
                return;
            }


            UserWithUserGroup connection = await _context.UserWithUserGroups
                .Include(p => p.GroupRole)
                .FirstOrDefaultAsync(
                    p => p.UserId == tuple.UserId && p.UserGroupId == tuple.GroupId);

            if (connection == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "connection not found")));
                return;
            }


            //... but the group needs at least one user with the creator role...
            bool? canLeave = await HasGroupAtLeastOnCreator(tuple.GroupId, tuple.UserId, connection.GroupRoleId);

            if (canLeave == null)
            {
                Console.WriteLine("ERROR system settings not found");
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.ServerError, "system setting not found")));
                return;
            }
            else if (canLeave.Value == false)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "cannot remove user, at least one creator must be in the group")));
                return;
            }

            SystemSetting systemSetting = await _context.SystemSettings
                .FirstOrDefaultAsync();

            //HasGroupAtLeastOnCreator checks the system settings...
            if (tuple.GroupId == systemSetting.DefaultUserGroupId)
            {
                //user cannot leave the default group...
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest,
                            "user cannot leave the default group")));
                return;
            }

            try
            {
                _context.UserWithUserGroups.Remove(connection);
                await _context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                await base.HandleDbError(ex);
                return;
            }

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "removed")));
        }

        /// <summary>
        /// returns the group role for the current user for the given group
        /// </summary>
        /// <param name="userGroupId"></param>
        /// <returns></returns>
        [HttpGet("role/for/{userGroupId}")]
        public async Task GetGroupRoleForGroup(int userGroupId)
        {
            if (!await base.IsLoggedIn()) return;

            int userId = GetUserId();

            UserGroup userGroup = await _context.UserGroups
                .FirstOrDefaultAsync(p => p.Id == userGroupId);


            if (userGroup == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user group not found")));
                return;
            }

            var connection = await _context.UserWithUserGroups
                .Include(p => p.GroupRole)
                .ThenInclude(p => p.GroupRolePermission)
                .FirstOrDefaultAsync(p => p.UserId == userId && p.UserGroupId == userGroupId);

            if (connection == null)
            {
                // await
                //     Response.WriteAsync(
                //         Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user is not part of the group")));
                // return;
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponseWithData<GroupRoleWithPermissionFullBase>(ResponseCode.Ok,
                            "user is not part of the group", null)));
                return;
            }

            if (connection.GroupRole == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "user has no group role")));
                return;
            }

            if (connection.GroupRole.GroupRolePermission == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "group role has no permission")));
                return;
            }

            var groupRole = connection.GroupRole;

            var permissionForUser = new GroupRoleWithPermissionFullBase
            (
                id: groupRole.Id,
                displayName: groupRole.DisplayName,
                canChangeOtherMembersRole: groupRole.GroupRolePermission.CanChangeOtherMembersRole,
                canCreateExercises: groupRole.GroupRolePermission.CanCreateExercises,
                canAddUserToGroup: groupRole.GroupRolePermission.CanAddUserToGroup,
                canAssessExercises: groupRole.GroupRolePermission.CanAssessExercises,
                canChangeExercises: groupRole.GroupRolePermission.CanChangeExercises,
                canDeleteExercises: groupRole.GroupRolePermission.CanDeleteExercises,
                canLockExercisesPermanently: connection.GroupRole.GroupRolePermission.CanLockExercisesPermanently,
                canManageExerciseReleases: groupRole.GroupRolePermission.CanManageExerciseReleases,
                canRemoveMemberFromGroup: groupRole.GroupRolePermission.CanRemoveMemberFromGroup,
                canSeeExercisesFromOthersInGroup: groupRole.GroupRolePermission.CanSeeExercisesFromOthersInGroup,
                canSeeOtherMembers: connection.GroupRole.GroupRolePermission.CanSeeOtherMembers,
                canChangeGroupData: connection.GroupRole.GroupRolePermission.CanChangeGroupData
            );

            await
                Response.WriteAsync(
                    Jc.Serialize(new BasicResponseWithData<GroupRoleWithPermissionFullBase>(ResponseCode.Ok,
                        "", permissionForUser)));
        }
    }

    /// <summary>
    /// represents a user group <see cref="UserGroup"/>
    /// </summary>
    public class UserGroupBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the display name
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the email
        /// </summary>
        public string Email { get; set; }        
    }

    public class UserGroupForBackend : UserGroupBase
    {
        
    }
    
    public class UserGroupFromBackend : UserGroupBase
    {
        /// <summary>
        /// the group role of the current user for the user group
        /// </summary>
        public GroupRoleWithPermissionFullBase GroupRole { get; set; }
    }
    
    public class UserGroupWithExerciseCountFromBackend : UserGroupFromBackend
    {
        /// <summary>
        /// the group role of the current user for the user group
        /// </summary>
        public GroupRoleWithPermissionFullBase GroupRole { get; set; }
        /**
         * number of exercises connected to the group
         */
        public int NumExercisesInGroup { get; set; }
    }

    /// <summary>
    /// used to change the group role of a member
    /// </summary>
    public class UserWithGroupRoleForBackend
    {
        public int UserId { get; set; }
        public int GroupId { get; set; }
        public int GroupRoleId { get; set; }
    }

    /// <summary>
    /// used to add users to a user group
    /// </summary>
    public class AddUsersToGroupTuplesForBackend
    {
        public int GroupId { get; set; }
        public List<int> UserIds { get; set; }
    }

    /// <summary>
    /// used to remove a member from a group
    /// </summary>
    public class RemoveUsersFromGroupTuplesForBackend
    {
        public int GroupId { get; set; }
        public int UserId { get; set; }
        public bool UseCurrentUser { get; set; }
    }

    /// <summary>
    /// describes a group member
    /// </summary>
    public class GroupUserFromBackend
    {
        public int Id { get; set; }

        /// <summary>
        /// the token of the user (id)
        /// </summary>
        public string Token { get; set; }

        /// <summary>
        /// the first name
        /// </summary>
        public string FirstName { get; set; }

        /// <summary>
        /// the last name
        /// </summary>
        public string LastName { get; set; }

        /// <summary>
        /// the email
        /// </summary>
        public string Email { get; set; }

        /// <summary>
        /// true: the user is activated, false: not
        /// </summary>
        public bool IsActivated { get; set; }

        //public bool IsCurrentUser { get; set; }

        /// <summary>
        /// the related group role id
        /// </summary>
        public int GroupRoleId { get; set; }

        /// <summary>
        /// true: this is the current user (e.g. cannot delete himself), false: not
        /// </summary>
        public bool IsCurrentUser { get; set; }
    }
}
