﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reactive.Linq;
using System.Threading;
using System.Threading.Tasks;
using ClientServer.Controllers.Core.Testing;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models.Exercises;
using ClientServer.Models.Exercises.Release;
using ClientServer.Models.Exercises.Solution;
using ClientServer.Models.Exercises.Tests;
using ClientServer.Models.Interfaces;
using ClientServer.Models.Users;
using Microsoft.EntityFrameworkCore;
using Newtonsoft.Json;

namespace ClientServer.Schedulers
{
    /// <summary>
    /// no longer a worker because assessment is only possible when a release gets unreleased... (manually or by the release scheduler)
    /// </summary>
    public static class SubmissionAssessmentWorker
    {
        /// <summary>
        /// the number of the run
        /// this will cycle from max int to min int which is ok as long as every request has a unique id (positive or negative)
        /// </summary>
        private static int runNumber = 0;

        /// <summary>
        /// true: an assessment is already running, then skip next iteration if needed,
        /// false: iteration/assessment can run
        /// 
        /// </summary>
        private static bool _assessmentRunning = false;

        /// <summary>
        /// an object for thread locking (we actually don't know if locking works with asp core but we try...)
        /// </summary>
        private static readonly Object Locker = new object();

        /// <summary>
        /// starts periodically checking for releases to be released or unrelesed (and releases/unreleses them)
        /// </summary>
        public static void Start()
        {
            Observable
                .Interval(TimeSpan.FromMinutes(1))
                //.Interval(TimeSpan.FromSeconds(20)) //for debug
                .Subscribe((min) => { NextRun(false); });
        }

        /// <summary>
        /// starts a new run/iteration in a new tread
        /// depending on Startup.RunAssessmentsAsync
        /// Startup.RunAssessmentsAsync = true 
        ///     run the assessment for the specific release in a new thread (asynchronously)
        /// 
        /// Startup.RunAssessmentsAsync = false
        ///     run the assessment for all closed releases that are not yet assessed (synchronously)
        /// </summary>
        public static void StartNextNewThread(int releaseId, bool reRunNotFinishedAssessments)
        {
            if (AppConfiguration.RunAssessmentsAsync)
            {
                //async, unsafe start a new thread and run the assessment...
                //the submit-test-server might reject tests because too many tests...
                Task.Factory.StartNew(() => { RunForRelease(releaseId); });
            }
            else
            {
                //sync, blocking with lock (but run in a new thread to not block the calling one (a user might closed the request manually...))
                Task.Factory.StartNew(() => NextRun(reRunNotFinishedAssessments));
            }
        }

        /// <summary>
        /// runs an assessment for a release
        /// </summary>
        /// <param name="releaseId"></param>
        [Obsolete("not properly implemented & unsafe")]
        public static void RunForRelease(int releaseId)
        {
            var optionsBuilder = new DbContextOptionsBuilder<YapexDbContext>();
            optionsBuilder.UseNpgsql(AppConfiguration.DbConnectionString);

            try
            {
                using (var context = new YapexDbContext(optionsBuilder.Options))
                {
                    var systemSettings = context.SystemSettings.FirstOrDefault();

                    if (systemSettings == null)
                    {
                        Console.WriteLine(
                            "ERROR system settings not found, cannot run submissions scheduler.RunForRelease");
                        return;
                    }

                    var testingController = new TestingController(context);

                    var finishedRelease = context.ExerciseReleases
                        .Where(
                            p =>
                                p.Id == releaseId &&
                                p.HasAutomaticAssessmentStarted == false && p.IsReleased == false &&
                                p.ReleasedAt != null
                                && p.Exercise.IsPermanentlyLocked == false)
                        //&& p.ShouldAutomaticAssessSubmissions // do not check for this because automatic assessments can only be started manually when the automatic assessment has finished...
                        .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                        .ThenInclude(p => p.Solutions)
                        //not include the whole graph here ... query every solution
                        .FirstOrDefault();

                    if (finishedRelease == null)
                    {
                        Console.WriteLine("ERROR: release with id '" + releaseId +
                                          "' not found but assessment was requested");
                        return;
                    }

                    if (finishedRelease.ShouldAutomaticAssessSubmissions)
                    {
                        var r = RunForRelease(finishedRelease, context, testingController, false, false, systemSettings)
                            .Result;
                    }
                    else
                    {
                        //we don't auto assess the solutions but for manual assessment we need an (empty) assessment

                        //add dummy releases to that we can manually assess
                        AddDummyAssessments(finishedRelease, context, testingController);

//                        finishedRelease.HasAutomaticAssessmentFinished = true;
//                        finishedRelease.HasAutomaticAssessmentStarted = true;
//                        finishedRelease.HadAutomaticAssessmentErrors = false;

                        try
                        {
                            context.SaveChanges();
                        }
                        catch (Exception ex)
                        {
                            Console.WriteLine("ERROR: " + ex);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR: " + ex);
            }
        }

        /// <summary>
        /// runs the next check for open releases and assesses them synchronously
        /// <param name="reRunNotFinishedAssessments">it is a good idea to call this func with this param set to true only on startup</param>
        /// </summary>
        public static void NextRun(bool reRunNotFinishedAssessments)
        {
            var now = DateTime.Now;

            if (_assessmentRunning)
            {
                return;
            }

            //ensure that only one assessment is running at the same time
            lock (Locker)
            {
                _assessmentRunning = true;
                if (AppConfiguration.IsDebugMode)
                {
                    //TODO maybe enable??
                    //Console.WriteLine("Assessment scheduler run # " + runNumber + " running at: " + now.ToLocalTime());
                }

                var optionsBuilder = new DbContextOptionsBuilder<YapexDbContext>();
                optionsBuilder.UseNpgsql(AppConfiguration.DbConnectionString);

                int assessedReleases = 0;


                try
                {
                    using (var context = new YapexDbContext(optionsBuilder.Options))
                    {
                        var systemSettings = context.SystemSettings.FirstOrDefault();

                        if (systemSettings == null)
                        {
                            Console.WriteLine(
                                "ERROR system settings not found, cannot run submissions scheduler.NextRun");
                            return;
                        }

                        var testingController = new TestingController(context);

                        List<ExerciseRelease> finishedReleases;

                        if (reRunNotFinishedAssessments)
                        {
                            finishedReleases = context.ExerciseReleases
                                .Where(
                                    p =>
                                        p.HasAutomaticAssessmentStarted == true //started
                                        && p.HasAutomaticAssessmentFinished == false //but not finished
                                        && p.IsReleased == false //users cannot longer edit solutions
                                        && p.ReleasedAt != null //must have been released at some point
                                        && p.Exercise.IsPermanentlyLocked == false
                                )
                                .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                                .ThenInclude(p => p.Solutions)
                                //not include the whole graph here ... query every solution
                                .ToList();
                        }
                        else
                        {
                            finishedReleases = context.ExerciseReleases
                                .Where(
                                    p =>
                                        p.HasAutomaticAssessmentStarted == false //not yet started
                                        && p.IsReleased == false //users cannot longer edit solutions
                                        && p.ReleasedAt != null //must have been released at some point
                                        && p.Exercise.IsPermanentlyLocked == false
                                )
                                .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                                .ThenInclude(p => p.Solutions)
                                //not include the whole graph here ... query every solution
                                .ToList();
                        }

                        assessedReleases = finishedReleases.Count;

                        foreach (var release in finishedReleases)
                        {
                            if (release.ShouldAutomaticAssessSubmissions)
                            {
                                var someError = RunForRelease(release, context, testingController,
                                    reRunNotFinishedAssessments, false, systemSettings).Result;
                            }
                            else
                            {
                                //add dummy releases to that we can manually assess
                                AddDummyAssessments(release, context, testingController);
                            }
                        }
                    }
                }
                catch (Exception ex)
                {
                    Console.WriteLine("ERROR: " + ex);
                }

                if (AppConfiguration.IsDebugMode && assessedReleases > 0)
                {
                    Console.WriteLine("Assessment scheduler run (NextRun) # " + runNumber + " finished at: " +
                                      now.ToLocalTime() +
                                      " # assessed releases: " + assessedReleases);
                }

                _assessmentRunning = false;
            }
        }


        /// <summary>
        /// adds dummy assessments for every solution, needed if we don't want to automatically assess the solutions but manually
        /// </summary>
        /// <param name="release"></param>
        /// <param name="context"></param>
        /// <param name="testingController"></param>
        /// <returns></returns>
        internal static bool AddDummyAssessments(ExerciseRelease release, YapexDbContext context,
            TestingController testingController)
        {
            ExerciseRelease finishedRelease = context.ExerciseReleases
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .ThenInclude(p => p.Solutions)
                    //not include the whole graph here ... query every solution
                    .FirstOrDefault(
                        p =>
                            p.Id == release.Id &&
//                            (p.HasAutomaticAssessmentStarted == false ||
//                             p.HasAutomaticAssessmentStarted && p.HasAutomaticAssessmentFinished) &&
                            p.IsReleased == false &&
                            p.ReleasedAt != null)
                ;

            if (finishedRelease == null)
            {
                Console.WriteLine("ERROR: requested release not found in addDummyAssessments");
                return false;
            }

            var solutionsIdsToAssess = new List<SolutionPrimaryKey>();

            #region lock release

            //first mark the release so that the next iteration won't includ this release
            release.HasAutomaticAssessmentStarted = true;
            //also reset from e.g. former runs
            finishedRelease.HadAutomaticAssessmentErrors = false;
            finishedRelease.HasAutomaticAssessmentFinished = false;
            try
            {
                context.SaveChanges();
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR: " + ex.Message);
                return false;
            }

            #endregion

            foreach (var participation in release.ExerciseReleaseWithUserAsParticipations)
            {
                //if the exercise was released for all plangs then we need to access solutions for every plang

                if (release.PLangId != null)
                {
                    //only 1 plang to assess
                    var solution = participation.Solutions.FirstOrDefault(p => p.PLangId == release.PLangId);

                    //an user could have a participation BUT NO solution!!! (release code entered but not entered)

                    if (solution == null)
                    {
                        //we cannot assess this...
                    }
                    else
                    {
                        solutionsIdsToAssess.Add(new SolutionPrimaryKey()
                        {
                            PLangId = solution.PLangId,
                            ExerciseReleaseId = solution.ExerciseReleaseId,
                            UserId = solution.UserId
                        });
                    }
                }
                else
                {
                    //solution for every plang...

                    foreach (var solution in participation.Solutions)
                    {
                        solutionsIdsToAssess.Add(new SolutionPrimaryKey()
                        {
                            PLangId = solution.PLangId,
                            ExerciseReleaseId = solution.ExerciseReleaseId,
                            UserId = solution.UserId
                        });
                    }
                }
            }

            List<Test> normalTests = new List<Test>();

            //only run normal tests if needed
            if (release.RunAlsoNormalTests)
            {
                normalTests =
                    context.Tests
                        .Include(p => p.TestSettings)
                        .Include(p => p.TestType)
                        .Where(p => p.ExerciseId == release.ExerciseId && p.IsSubmitTest == false)
                        .ToList();
            }


            var submitTests =
                context.Tests
                    .Include(p => p.TestSettings)
                    .Include(p => p.TestType)
                    .Where(p => p.ExerciseId == release.ExerciseId && p.IsSubmitTest == true)
                    .ToList();

            int maxNormalTestPoints = normalTests.Sum(p => p.Weight);
            int maxSubmitTestPoints = submitTests.Sum(p => p.Weight);
            bool runSuccess = false;
            bool someError = false;

            //now we have all solutions that needs to be assessed
            foreach (var solutionPrimaryKey in solutionsIdsToAssess)
            {
                Solution solution = context.Solutions
                    .Include(p => p.PLang)
                    .Include(p => p.Assessment)
//                .Include(p => p.SolutionVersions)
                    .Include(p => p.TestResults)
//                .Include(p => p.SolutionVersions)
                    .Include(p => p.SolutionFiles)
                    .FirstOrDefault(
                        p => p.PLangId == solutionPrimaryKey.PLangId && p.UserId == solutionPrimaryKey.UserId
                                                                     && p.ExerciseReleaseId ==
                                                                     solutionPrimaryKey.ExerciseReleaseId);

//                SolutionVersion latestSolutionVersion = solution.SolutionVersions.OrderBy(p => p.CreatedAt).LastOrDefault();

                if (solution == null)
                {
                    Console.WriteLine("[WARNING] solution not found (skipping), cannot access, userId: " +
                                      solutionPrimaryKey.UserId +
                                      " releaseId: " + solutionPrimaryKey.ExerciseReleaseId + " plangId: " +
                                      solutionPrimaryKey.PLangId);
                    continue;
                }

                if (solution.Assessment == null)
                {
                    solution.Assessment = new SolutionAssessment()
                    {
                        MaxSubmitTestPoints = maxSubmitTestPoints,
                        MaxNormalTestPoints = maxNormalTestPoints,
                        ManualPoints = null,
                        Solution = solution,
                        //SolutionVersion = null,
                        SubmitTestPoints = null,
                        NormalTestPoints = null,
//                        SolutionVersion = latestSolutionVersion,
                        FeedbackForStudent = "",
                        NoteForOtherTutors = "",
                        LastAssessmentErrorMessage = null
                    };
                }
                else
                {
                    //rerun ... reset props
                    solution.Assessment.MaxSubmitTestPoints = maxSubmitTestPoints;
                    solution.Assessment.MaxNormalTestPoints = maxNormalTestPoints;
                    solution.Assessment.ManualPoints = null;
                    solution.Assessment.Solution = solution;
//                    solution.Assessment.SolutionVersion = null;
                    solution.Assessment.SubmitTestPoints = null;
                    solution.Assessment.NormalTestPoints = null;
//                    solution.Assessment.SolutionVersion = latestSolutionVersion;
                    solution.Assessment.FeedbackForStudent = solution.Assessment.FeedbackForStudent;
                    solution.Assessment.NoteForOtherTutors = solution.Assessment.NoteForOtherTutors;
                    solution.Assessment.LastAssessmentErrorMessage = null;
                }
            }

            finishedRelease.HadAutomaticAssessmentErrors = someError;
            finishedRelease.HasAutomaticAssessmentFinished = true;

            try
            {
                context.SaveChanges();
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR when saving dummy automatic assessment: " + ex.Message);
                return false;
            }

            if (someError)
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// runs the automatic assessment for the given release (release will be queried --> release do not need something attached (navigation properties))
        /// ONLY IF THE RELEASE IS NOT RELEASED, HAS A RELEASED AT DATE, HasAutomaticAssessmentStarted = FALSE or has started and finished
        /// </summary>
        /// <param name="release">the release to use</param>
        /// <param name="context">the db context</param>
        /// <param name="testingController">the testing controller to run the tests</param>
        /// <param name="resetTestResultsFirst">true: reset all test results before running the assessment, false: not</param>
        /// <returns>true: ok, false: no matching release found</returns>
        internal static async Task<bool> RunForRelease(ExerciseRelease release, YapexDbContext context,
            TestingController testingController, bool reRunNotFinishedAssessments, bool resetTestResultsFirst,
            SystemSetting systemSettings)
        {
            ExerciseRelease finishedRelease = await context.ExerciseReleases
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .ThenInclude(p => p.Solutions)
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .ThenInclude(p => p.User)
                    .Include(p => p.Exercise.CodeTemplates)
                    .ThenInclude(p => p.TemplateFiles)
                    //not include the whole graph here ... query every solution
                    .FirstOrDefaultAsync(
                        p =>
                            p.Id == release.Id &&
                            (reRunNotFinishedAssessments && p.HasAutomaticAssessmentStarted == true &&
                             p.HasAutomaticAssessmentFinished == false ||
                             p.HasAutomaticAssessmentStarted && p.HasAutomaticAssessmentFinished ||
                             p.HasAutomaticAssessmentStarted == false && p.HasAutomaticAssessmentFinished == false
                            ) &&
                            p.IsReleased == false &&
                            p.ReleasedAt != null
//                            && p.ShouldAutomaticAssessSubmissions
                    )
                ;

            if (finishedRelease == null)
            {
                Console.WriteLine("ERROR: requested release not found in RunForRelease");
                return false;
            }

            var solutionsIdsToAssess = new List<SolutionPrimaryKey>();

            #region lock release

            //first mark the release so that the next iteration won't includ this release
            release.HasAutomaticAssessmentStarted = true;
            //also reset from e.g. former runs
            finishedRelease.HadAutomaticAssessmentErrors = false;
            finishedRelease.HasAutomaticAssessmentFinished = false;
            try
            {
                await context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR: " + ex.Message);
                return false;
            }

            #endregion

            foreach (var participation in release.ExerciseReleaseWithUserAsParticipations)
            {
                //solution for every plang... we should only accepted solutions with the right plangs
                //so accept every plang here

                foreach (var solution in participation.Solutions)
                {
                    solutionsIdsToAssess.Add(new SolutionPrimaryKey()
                    {
                        PLangId = solution.PLangId,
                        ExerciseReleaseId = solution.ExerciseReleaseId,
                        UserId = solution.UserId,
                        UserLastName = participation.User.LastName
                    });
                }
            }

            List<Test> normalTests = new List<Test>();

            normalTests =
                await context.Tests
                    .Include(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceTestAsset)
                    .Include(p => p.TestSettings)
                    .Include(p => p.TestType)
                    .Where(p => p.ExerciseId == release.ExerciseId && p.IsSubmitTest == false)
                    .ToListAsync();


            var submitTests =
                await context.Tests
                    .Include(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceTestAsset)
                    .Include(p => p.TestSettings)
                    .Include(p => p.TestType)
                    .Where(p => p.ExerciseId == release.ExerciseId && p.IsSubmitTest == true)
                    .ToListAsync();

            int maxNormalTestPoints = normalTests.Sum(p => p.Weight);
            int maxSubmitTestPoints = submitTests.Sum(p => p.Weight);
            bool runSuccess = false;
            bool someError = false;

            //only run normal tests if needed
            if (release.RunAlsoNormalTests == false)
            {
                normalTests = new List<Test>();
            }
            
            //get all test file contents
            var normalTestFilesIdsToRead = normalTests
                    .SelectMany(p => p.AssetReferences)
                    .Select(p => p.FileReferenceTestAssetId)
                    .ToList()
                ;

            var submitTestFilesIdsToRead = submitTests
                    .SelectMany(p => p.AssetReferences)
                    .Select(p => p.FileReferenceTestAssetId)
                    .ToList()
                ;

            var testFilesIdsToRead = normalTestFilesIdsToRead.Concat(submitTestFilesIdsToRead).ToArray();

            var testAssetsBasePath = Files.GetUploadFilePath(UploadDirType.TestAssets);

            var testFilesDict = await Files.ReadUploadedFilesAsArray(testAssetsBasePath, testFilesIdsToRead);


            if (resetTestResultsFirst)
            {
                List<Solution> allSolutions = await context.Solutions
                        .Include(p => p.Assessment)
                        .Where(p => p.ExerciseReleaseId == release.Id)
                        .ToListAsync()
                    ;

                foreach (var solution in allSolutions)
                {
                    if (solution.Assessment == null)
                    {
                        //nothing to reset
                    }
                    else
                    {
                        solution.Assessment.MaxSubmitTestPoints = maxSubmitTestPoints;
                        solution.Assessment.MaxNormalTestPoints = maxNormalTestPoints;
                        solution.Assessment.ManualPoints = solution.Assessment.ManualPoints;
                        solution.Assessment.Solution = solution;
                        solution.Assessment.SubmitTestPoints = null;
                        solution.Assessment.NormalTestPoints = null;
                        solution.Assessment.FeedbackForStudent = solution.Assessment.FeedbackForStudent;
                        solution.Assessment.NoteForOtherTutors = solution.Assessment.NoteForOtherTutors;
                        solution.Assessment.LastAssessmentErrorMessage = null;
                    }
                }

                try
                {
                    await context.SaveChangesAsync(); //save so frontend can see that we reset the assessments
                }
                catch (Exception ex)
                {
                    Console.WriteLine("ERROR: " + ex.Message);
                    return false;
                }
            }

            var sortedSolutionIdsToAssess = solutionsIdsToAssess.OrderBy(p => p.UserLastName)
                .ToList();

            //now we have all solutions that needs to be assessed
            foreach (var solutionPrimaryKey in sortedSolutionIdsToAssess)
            {
                var codeTemplate = finishedRelease.Exercise.CodeTemplates
                        .FirstOrDefault(p => p.PLangId == solutionPrimaryKey.PLangId)
                    ;

                if (codeTemplate == null)
                {
                    Console.WriteLine("ERROR could not find code template, userId:" +
                                      solutionPrimaryKey.UserId + " release id: " +
                                      solutionPrimaryKey.ExerciseReleaseId +
                                      " pLangId: " + solutionPrimaryKey.PLangId);
                    runSuccess = false;
                }
                else
                {
                    runSuccess = await RunAutomaticAssessment(solutionPrimaryKey, maxNormalTestPoints,
                        maxSubmitTestPoints,
                        normalTests, submitTests, context, codeTemplate, testingController, systemSettings, testFilesDict);
                }


                if (runSuccess == false)
                {
                    Console.WriteLine("WARNING did not run all assessments because some assessment failed, userId:" +
                                      solutionPrimaryKey.UserId + " release id: " +
                                      solutionPrimaryKey.ExerciseReleaseId +
                                      " pLangId: " + solutionPrimaryKey.PLangId + ", releaseCode: " +
                                      release.GeneratedCode);
                    someError = true;

                    if (Constants.BreakAutoAssessmentAfterFirstError)
                    {
                        //don't run the others... if an assessment "destroyed" the test server then we can't run the others
                        //if the test server is not reachable... we can't run the others
                        //it's unlikely that only one solution will throw somehow and the other don't
                        //  so if there is an error we can see it (because not all are assessed in the overview)

                        break;
                    }
                }
            }

            finishedRelease.HadAutomaticAssessmentErrors = someError;
            finishedRelease.HasAutomaticAssessmentFinished = true;

            try
            {
                await context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR when saving finished automatic assessment: " + ex.Message);
                return false;
            }

            if (someError)
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// runs the automatic assessment for a single user
        ///
        /// not called in this scheduler
        /// </summary>
        /// <param name="release">the release</param>
        /// <param name="pk">the primary key of the solution to use</param>
        /// <param name="context">the db context</param>
        /// <param name="testingController">the testing controller</param>
        /// <returns></returns>
        internal static async Task<bool> RunSingleAutomaticAssessment(ExerciseRelease release, SolutionPrimaryKey pk,
            YapexDbContext context, TestingController testingController, SystemSetting systemSettings)
        {
            ExerciseRelease finishedRelease = await context.ExerciseReleases
                    .Include(p => p.ExerciseReleaseWithUserAsParticipations)
                    .ThenInclude(p => p.Solutions)
                    .Include(p => p.Exercise.CodeTemplates)
                    .ThenInclude(p => p.TemplateFiles)
                    //not include the whole graph here ... query every solution
                    .FirstOrDefaultAsync(
                        p =>
                            p.Id == release.Id &&
                            p.HasAutomaticAssessmentStarted &&
                            p.HasAutomaticAssessmentFinished && //only allowed when already ran
                            //the automatic test must/should run first...
                            p.IsReleased == false &&
                            p.ReleasedAt != null //was released at some time
                    )
                ;

            if (finishedRelease == null)
            {
                return false;
            }

            List<Test> normalTests = new List<Test>();
            int maxNormalTestPoints = -1;
            if (release.RunAlsoNormalTests)
            {
                normalTests =
                    await context.Tests
                        .Include(p => p.AssetReferences)
                        .ThenInclude(p => p.FileReferenceTestAsset)
                        .Include(p => p.TestSettings)
                        .Include(p => p.TestType)
                        .Where(p => p.ExerciseId == release.ExerciseId && p.IsSubmitTest == false)
                        .ToListAsync();

                maxNormalTestPoints = normalTests.Sum(p => p.Weight);
            }

            var submitTests =
                await context.Tests
                    .Include(p => p.AssetReferences)
                    .ThenInclude(p => p.FileReferenceTestAsset)
                    .Include(p => p.TestSettings)
                    .Include(p => p.TestType)
                    .Where(p => p.ExerciseId == release.ExerciseId && p.IsSubmitTest == true)
                    .ToListAsync();


            int maxSubmitTestPoints = submitTests.Sum(p => p.Weight);
            bool runSuccess = false;

            
            //get all test file contents
            var normalTestFilesIdsToRead = normalTests
                    .SelectMany(p => p.AssetReferences)
                    .Select(p => p.FileReferenceTestAssetId)
                    .ToList()
                ;

            var submitTestFilesIdsToRead = submitTests
                    .SelectMany(p => p.AssetReferences)
                    .Select(p => p.FileReferenceTestAssetId)
                    .ToList()
                ;

            var testFilesIdsToRead = normalTestFilesIdsToRead.Concat(submitTestFilesIdsToRead).ToArray();

            var testAssetsBasePath = Files.GetUploadFilePath(UploadDirType.TestAssets);

            var testFilesDict = await Files.ReadUploadedFilesAsArray(testAssetsBasePath, testFilesIdsToRead);

            //now we have all solutions that needs to be assessed

            var codeTemplate = finishedRelease.Exercise.CodeTemplates
                    .FirstOrDefault(p => p.PLangId == pk.PLangId)
                ;

            if (codeTemplate == null)
            {
                Console.WriteLine("[ERROR] code template not found, userId: " + pk.UserId +
                                  " releaseId: " + pk.ExerciseReleaseId + " plangId: " + pk.PLangId);
                runSuccess = false;
            }
            else
            {
                runSuccess = await RunAutomaticAssessment(pk, maxNormalTestPoints, maxSubmitTestPoints,
                    normalTests, submitTests, context, codeTemplate, testingController, systemSettings, testFilesDict);
            }


            try
            {
                await context.SaveChangesAsync();
            }
            catch (Exception ex)
            {
                Console.WriteLine("ERROR when saving (single) finished automatic assessment: " + ex.Message);
                return false;
            }

            if (!runSuccess)
            {
                return false;
            }

            return true;
        }

        /// <summary>
        /// runs the automatic assessment for the given solution (by primary key)
        /// if the normal test should not be run then the normal tests are just empty and maxNormalTestPoints is -1
        ///
        /// THE CALLING method must ensure that testAssetFileContents contains all an entry for every test asset (null is not allowed and will print an warning and return false)
        /// </summary>
        /// <param name="pk">the primary key of the solution</param>
        /// <param name="maxSubmitTestPoints">the max test points from the submit tests</param>
        /// <param name="normalTests">the normal tests</param>
        /// <param name="submitTests">the submission tests</param>
        /// <param name="context">the db context</param>
        /// <param name="maxNormalTestPoints">the max test points from the normal tests</param>
        /// <param name="codeTemplate">the code template so we can replace the readonly file content, also make sure the files are loadedd, MUST NOT BE NULL</param>
        /// <param name="testingController">the testing controller to run the tests</param>
        /// <param name="systemSettings"></param>
        /// <param name="testAssetFileContents">for all tests (normal and submit) the test file contents</param>
        /// <returns>true: everything ok, false: some error</returns>
        private static async Task<bool> RunAutomaticAssessment(SolutionPrimaryKey pk, int maxNormalTestPoints,
            int maxSubmitTestPoints,
            List<Test> normalTests, List<Test> submitTests, YapexDbContext context,
            CodeTemplate codeTemplate,
            TestingController testingController,
            SystemSetting systemSettings,
            Dictionary<int, byte[]> testAssetFileContents
            )
        {
            //runNumber is not really a valid request distinction value because when we run async assessments
            // ++ ist not atomic because it's translated to i += 1 or something ... so >1 assessments might get the same number here...
            //the test server, however, uses microtime so this should be enough

            //so we use  Interlocked.Increment to do the same as runNumber++ but threadsafe!
            Interlocked.Increment(ref runNumber);

            Solution solution = await context.Solutions
                .Include(p => p.PLang)
                .Include(p => p.Assessment)
                .Include(p => p.TestResults)
                .Include(p => p.SolutionFiles)
                .FirstOrDefaultAsync(
                    p => p.PLangId == pk.PLangId && p.UserId == pk.UserId &&
                         p.ExerciseReleaseId == pk.ExerciseReleaseId);


            if (solution == null)
            {
                Console.WriteLine("[ERROR] solution not found (skipping), cannot access, userId: " + pk.UserId +
                                  " releaseId: " + pk.ExerciseReleaseId + " plangId: " + pk.PLangId);
                return false;
            }

            if (codeTemplate == null)
            {
                Console.WriteLine("[ERROR] code template was null (skipping), userId: " + pk.UserId +
                                  " releaseId: " + pk.ExerciseReleaseId + " plangId: " + pk.PLangId);
                return false;
            }

            //make sure readonly files are replaced with the up-to-date versions
            var errorMsg = UserSolutionHelper.ReplaceReadonlyFiles(solution, codeTemplate, false, true);

            if (errorMsg != null)
            {
                return false;
            }

            int? normalPoints = null;
            int? submitPoints = null;

            if (solution.SolutionFiles.Count == 0)
            {
                //normally this cannot happen because we ensure if the user accesses the exercise he/she gets the default template as solution (with solution version)

                if (solution.Assessment == null)
                {
                    solution.Assessment = new SolutionAssessment()
                    {
                        MaxSubmitTestPoints = maxSubmitTestPoints,
                        MaxNormalTestPoints = maxNormalTestPoints,
                        ManualPoints = null,
                        Solution = solution,
                        SubmitTestPoints = null,
                        NormalTestPoints = null,
                        FeedbackForStudent = "",
                        NoteForOtherTutors = "",
                        LastAssessmentErrorMessage = null,
                    };
                }
                else
                {
                    solution.Assessment.MaxSubmitTestPoints = maxSubmitTestPoints;
                    solution.Assessment.MaxNormalTestPoints = maxNormalTestPoints;
                    solution.Assessment.ManualPoints = solution.Assessment.ManualPoints;
                    solution.Assessment.Solution = solution;
                    solution.Assessment.SubmitTestPoints = null;
                    solution.Assessment.NormalTestPoints = null;
                    solution.Assessment.FeedbackForStudent = solution.Assessment.FeedbackForStudent;
                    solution.Assessment.NoteForOtherTutors = solution.Assessment.NoteForOtherTutors;
                    solution.Assessment.LastAssessmentErrorMessage = null;
                }
            }
            else
            {
                //don't reset assessment so we can preserve the comments (tutor feedback...)

                #region test the solution

                var hasError = false;
                string errorMessageForFrontend = null;

                #region normal tests

                List<AllTest> allNormalTests;

                try
                {
                    allNormalTests = normalTests.Select(p => AllTest.CreateFrom(p, testAssetFileContents)).ToList();
                }
                catch (Exception e)
                {
                    Console.WriteLine("[ERROR]: " + e);
                    return false;
                }
                
                Tuple<ResponseFromTestServer, string> normalAnswersFromTestServer = null;


                //limit the tests run with one request to the specified max... hopefully this is exactly the amount the test server accepts
                var partedTests = new List<List<AllTest>>();

                List<AllTest> takenTests;
                int takenTestsCount = 0;

                while (takenTestsCount < allNormalTests.Count)
                {
                    takenTests = new List<AllTest>();
                    //even though these are normal tests ... we donign assessment so use the submit test server and also the
                    //MaxNumberOfTestsWithOneRequestSubmitTestServer as max number of tests for 1 request
                    for (int i = 0;
                        i < systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer &&
                        takenTestsCount < allNormalTests.Count;
                        i++)
                    {
                        takenTests.Add(allNormalTests[takenTestsCount]);
                        takenTestsCount++;
                    }

                    if (takenTests.Count > 0)
                    {
                        partedTests.Add(takenTests);
                    }
                }

                //run in parts so the test server is not constantly occupied
                foreach (var testParts in partedTests)
                {
                    try
                    {
                        normalAnswersFromTestServer = await testingController._RunTestOnTestServer(solution,
                            testParts,
                            solution.PLang,
                            systemSettings.SubmitTestServerUrl, true, runNumber,
                            systemSettings); //we assess so always use the submit test server
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("ERROR running normal test: " + ex.Message);
                        hasError = true;
                        errorMessageForFrontend = ex.Message;
                        goto saveTestResults;
                        return false;
                    }

                    if (normalAnswersFromTestServer?.Item1?.Data == null)
                    {
                        //something went wrong... probably the test server is offline...no need to test others
                        hasError = true;
                        errorMessageForFrontend = "unknown error";

                        if (normalAnswersFromTestServer != null && normalAnswersFromTestServer.Item1 == null)
                            errorMessageForFrontend = normalAnswersFromTestServer.Item2;


                        goto saveTestResults;
                    }
                    else
                    {
                        foreach (var normalTest in normalTests)
                        {
                            //find the result for the normal test

                            AnswerTestFromTestServer testResult =
                                normalAnswersFromTestServer.Item1.Data.FirstOrDefault(p => p.TestId == normalTest.Id);

                            if (testResult == null)
                            {
                                //no result...
                                continue;
                            }

                            if (testResult.Passed)
                            {
                                if (normalPoints == null) normalPoints = 0;
                                normalPoints += normalTest.Weight;
                            }
                            else
                            {
                                //we have ran at least one test
                                if (normalPoints == null) normalPoints = 0;
                            }

                            //set the test result for the normal tests (it does not matter if the user can see this)

                            var oldNormalTestResult = solution.TestResults.FirstOrDefault(
                                p => p.TestId == normalTest.Id);

                            if (oldNormalTestResult == null)
                            {
                                oldNormalTestResult = new TestWithSingleSolutionAsTestResult()
                                {
                                    HasCompiled = testResult.HasCompiled,
                                    Passed = testResult.Passed,
                                    ProgramExitCode = testResult.ProgramExitCode,
                                    TestResultCode = testResult.TestResultCode,
                                    Protocol = testResult.Protocol,
                                    Test = normalTest,
                                    Solution = solution
                                };
                                solution.TestResults.Add(oldNormalTestResult);
                            }
                            else
                            {
                                oldNormalTestResult.HasCompiled = testResult.HasCompiled;
                                oldNormalTestResult.Passed = testResult.Passed;
                                oldNormalTestResult.ProgramExitCode = testResult.ProgramExitCode;
                                oldNormalTestResult.TestResultCode = testResult.TestResultCode;
                                oldNormalTestResult.Protocol = testResult.Protocol;
                            }


                            //save for safety
                            try
                            {
                                await context.SaveChangesAsync();
                            }
                            catch (Exception ex)
                            {
                                Console.WriteLine("ERROR saving test result: " + ex.Message);
                                //stop tests because others will normally fail too
                                hasError = true;
                                errorMessageForFrontend = "error saving test result to db";
                                goto saveTestResults;
                                return false;
                            }
                        }
                    }
                }

                #endregion

                #region submit tests

                List<AllTest> allSubmitTests;

                try
                {
                    allSubmitTests = submitTests.Select(p => AllTest.CreateFrom(p, testAssetFileContents)).ToList();
                }
                catch (Exception e)
                {
                    Console.WriteLine(e);
                    return false;
                }
                
                Tuple<ResponseFromTestServer, string> submitAnswersFromTestServer = null;


                //limit the tests run with one request to the specified max... hopefully this is exactly the amount the test server accepts
                partedTests = new List<List<AllTest>>();

                takenTests = null;
                takenTestsCount = 0;

                while (takenTestsCount < allSubmitTests.Count)
                {
                    takenTests = new List<AllTest>();
                    for (int i = 0;
                        i < systemSettings.MaxNumberOfTestsWithOneRequestSubmitTestServer &&
                        takenTestsCount < allSubmitTests.Count;
                        i++)
                    {
                        takenTests.Add(allSubmitTests[takenTestsCount]);
                        takenTestsCount++;
                    }

                    if (takenTests.Count > 0)
                    {
                        partedTests.Add(takenTests);
                    }
                }


                foreach (var testParts in partedTests)
                {
                    try
                    {
                        submitAnswersFromTestServer = await testingController._RunTestOnTestServer(solution,
                            testParts,
                            solution.PLang,
                            systemSettings.SubmitTestServerUrl, true, runNumber, systemSettings);
                    }
                    catch (Exception ex)
                    {
                        Console.WriteLine("ERROR running submit test: " + ex.Message);
                        hasError = true;
                        errorMessageForFrontend = ex.Message;
                        goto saveTestResults;
                        return false;
                    }


                    if (submitAnswersFromTestServer?.Item1?.Data == null)
                    {
                        //something went wrong... probably the test server is offline...no need to test others
                        hasError = true;

                        if (submitAnswersFromTestServer != null && submitAnswersFromTestServer.Item1 == null)
                            errorMessageForFrontend = submitAnswersFromTestServer.Item2;

                        goto saveTestResults;
                    }
                    else
                    {
                        foreach (var submitTest in submitTests)
                        {
                            AnswerTestFromTestServer testResult =
                                submitAnswersFromTestServer.Item1.Data.FirstOrDefault(p => p.TestId == submitTest.Id);

                            if (testResult == null)
                            {
                                //no result...
                                continue;
                            }

                            if (testResult.Passed)
                            {
                                if (submitPoints == null) submitPoints = 0;
                                submitPoints += submitTest.Weight;
                            }
                            else
                            {
                                //we have ran at least one test
                                if (submitPoints == null) submitPoints = 0;
                            }

                            //but set the test result in the db for the submit tests

                            var oldSubmitTestResult = solution.TestResults.FirstOrDefault(
                                p => p.TestId == submitTest.Id);

                            if (oldSubmitTestResult == null)
                            {
                                oldSubmitTestResult = new TestWithSingleSolutionAsTestResult()
                                {
                                    HasCompiled = testResult.HasCompiled,
                                    Passed = testResult.Passed,
                                    ProgramExitCode = testResult.ProgramExitCode,
                                    TestResultCode = testResult.TestResultCode,
                                    Protocol = testResult.Protocol,
                                    Test = submitTest,
                                    Solution = solution
                                };
                                solution.TestResults.Add(oldSubmitTestResult);
                            }
                            else
                            {
                                oldSubmitTestResult.HasCompiled = testResult.HasCompiled;
                                oldSubmitTestResult.Passed = testResult.Passed;
                                oldSubmitTestResult.ProgramExitCode = testResult.ProgramExitCode;
                                oldSubmitTestResult.TestResultCode = testResult.TestResultCode;
                                oldSubmitTestResult.Protocol = testResult.Protocol;
                            }


                            //save for safety
                            try
                            {
                                await context.SaveChangesAsync();
                            }
                            catch (Exception ex)
                            {
                                Console.WriteLine("ERROR saving submit test result: " + ex.Message);
                                //stop tests because others will normally fail too
                                hasError = true;
                                errorMessageForFrontend = "error saving submit test result to db";
                                goto saveTestResults;
                                return false;
                            }
                        }
                    }
                }

                #endregion

                #endregion

                saveTestResults:
                //after all tests have run ... save assessment result

                if (solution.Assessment == null)
                {
                    solution.Assessment = new SolutionAssessment()
                    {
                        MaxSubmitTestPoints = maxSubmitTestPoints,
                        MaxNormalTestPoints = maxNormalTestPoints,
                        ManualPoints = null,
                        Solution = solution,
                        SubmitTestPoints = hasError ? null : (int?) submitPoints,
                        NormalTestPoints = hasError ? null : (int?) normalPoints,
                        FeedbackForStudent = "",
                        NoteForOtherTutors = "",
                        LastAssessmentErrorMessage = null
                    };
                }
                else
                {
                    solution.Assessment.MaxSubmitTestPoints = maxSubmitTestPoints;
                    solution.Assessment.MaxNormalTestPoints = maxNormalTestPoints;
                    solution.Assessment.ManualPoints = solution.Assessment.ManualPoints;
                    solution.Assessment.Solution = solution;
                    solution.Assessment.SubmitTestPoints = hasError ? null : (int?) submitPoints;
                    solution.Assessment.NormalTestPoints = hasError ? null : (int?) normalPoints;
                    solution.Assessment.FeedbackForStudent = solution.Assessment.FeedbackForStudent;
                    solution.Assessment.NoteForOtherTutors = solution.Assessment.NoteForOtherTutors;
                    solution.Assessment.LastAssessmentErrorMessage = hasError ? errorMessageForFrontend : null;
                }


                try
                {
                    await context.SaveChangesAsync();
                }
                catch (Exception ex)
                {
                    Console.WriteLine("ERROR saving automatic assessment result: " + ex.Message);
                    hasError = true;
                    return false;
                }

                return !hasError;
            }


            return true;
        }
    }

    /// <summary>
    /// the composite primary for a solution
    /// </summary>
    public class SolutionPrimaryKey
    {
        public int ExerciseReleaseId { get; set; }
        public int UserId { get; set; }
        public int PLangId { get; set; }

        /// <summary>
        /// used to sort the solutions before assessment e.g. to see where we stopped
        /// </summary>
        public string UserLastName { get; set; }
    }
}
